"""
Module to provide constants to deal with extensions.
"""

# pylint: disable=too-few-public-methods


class ExtensionManagerConstants:
    """
    Class to provide constants to deal with extensions.
    """

    EXTENSION_VERSION_NOT_IMPLEMENTED = "0.0.0"

    EXTENSION_INTERFACE_VERSION_BASIC = 1


# pylint: enable=too-few-public-methods
